"""Init-file for scope as a python-package."""
from pysignalscope.scope import *
from pysignalscope.functions import *
from pysignalscope.colors import *
from pysignalscope.generalplotsettings import *
from pysignalscope.impedance import *
from pysignalscope.impedance_dataclass import *
